@@ -4,11 +4,15 @@ import click
 import distro
 import PyInquirer
 
+from . import log
 from . import core
 from . import util
 from . import config
 
 
+logger = log.AppmanLogger(__file__, "DEBUG", "DEBUG")
+
+
 class BaseCommand(click.Command):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -27,6 +31,12 @@ class BaseCommand(click.Command):
             click.Option(
                 ("--os", "-os"), help="If no OS given, will use current detected OS"
             ),
+            click.Option(
+                ("--quiet", "-q"), is_flag=True, help="Print minimum information"
+            ),
+            click.Option(
+                ("--verbose", "-v"), is_flag=True, help="Print debugging information"
+            ),
         ] + self.params
 
     @staticmethod
@@ -36,14 +46,14 @@ class BaseCommand(click.Command):
 
 
 @click.group()
-@click.option("--verbose", "-v", is_flag=True, help="Print error stacktrace")
 @click.version_option(message="%(prog)s %(version)s")
 @click.pass_context
-def cli(ctx, verbose):
+def cli(ctx):
+    """Cross-platform application management aggregator"""
     try:
         os = platform.system().lower()
         if os not in config.OS_SUPPORTED:
-            util.print_error(f"{os} is not supported")
+            logger.error(f"{os} is not supported")
             return
 
         if os == "linux":
@@ -51,81 +61,81 @@ def cli(ctx, verbose):
 
         am = core.AppMan(config.BUCKET_PKG)
         am.load_user_data(config.USER_DATA_PKG)
-        ctx.obj = {"appman": am, "os": os, "verbose": verbose}
+        ctx.obj = {"appman": am, "os": os}
     except Exception as e:
-        e.verbose = verbose
+        e.verbose = True
         raise
 
 
-@cli.command(cls=BaseCommand)
+@cli.command(cls=BaseCommand, short_help="Install packages in user list")
 @click.option("--test", "-t", is_flag=True, help="Print commands without running")
 @click.pass_context
-def install(ctx, package_id, package_type, labels, os, test):
-    verbose = ctx.obj["verbose"]
-    try:
-        if not package_type:
-            package_type = prompt_package_type()
-
-        run_command(ctx, "install", package_id, package_type, labels, os, test, verbose)
-    except Exception as e:
-        e.verbose = verbose
-        raise
-
-
-@cli.command(cls=BaseCommand)
-@click.option("--test", "-t", is_flag=True, help="Print commands without running")
-@click.pass_context
-def uninstall(ctx, package_id, package_type, labels, os, test):
-    verbose = ctx.obj["verbose"]
-
+def install(ctx, package_id, package_type, labels, os, verbose, quiet, test):
+    """Install packages in user list"""
     try:
         if not package_type:
             package_type = prompt_package_type()
 
         run_command(
-            ctx, "uninstall", package_id, package_type, labels, os, test, verbose
+            ctx, "install", package_id, package_type, labels, os, test, verbose, quiet
         )
     except Exception as e:
         e.verbose = verbose
         raise
 
 
-@cli.command(cls=BaseCommand)
+@cli.command(cls=BaseCommand, short_help="Uninstall packages in user list")
+@click.option("--test", "-t", is_flag=True, help="Print commands without running")
 @click.pass_context
-def search(ctx, package_type, package_id, labels, os):
-    os = os or ctx.obj["os"]
-    appman = ctx.obj["appman"]
-    verbose = ctx.obj["verbose"]
-
+def uninstall(ctx, package_id, package_type, labels, os, verbose, quiet, test):
+    """Uninstall packages in user list"""
     try:
         if not package_type:
             package_type = prompt_package_type()
 
-        if package_id:
-            pkg = appman.get_package(package_type, package_id)
-            if pkg:
-                util.print_info(f"Package definition for '{package_id}' found")
-            else:
-                util.print_info(f"Package definition for '{package_id}' not found")
-            return
-
-        pkgs = appman.get_packages(package_type, os, labels=labels)
-        if not pkgs:
-            util.print_info("No packages found")
-            return
-
-        for pkg in pkgs:
-            util.print_info(pkg.id)
+        run_command(
+            ctx, "uninstall", package_id, package_type, labels, os, test, verbose, quiet
+        )
     except Exception as e:
         e.verbose = verbose
         raise
 
 
-@cli.command(cls=BaseCommand)
+@cli.command(cls=BaseCommand, short_help="Search for packages available in bucket")
 @click.pass_context
-def list(ctx, package_type, package_id, labels, os):
+def search(ctx, package_type, package_id, labels, os, verbose, quiet):
+    """Search for packages available in bucket"""
+    os = os or ctx.obj["os"]
+    appman = ctx.obj["appman"]
+
+    try:
+        if not package_type:
+            package_type = prompt_package_type()
+        if package_id:
+            pkg = appman.get_package(package_type, package_id)
+            if pkg:
+                logger.console(f"Package definition for '{package_id}' found")
+            else:
+                logger.console(f"Package definition for '{package_id}' not found")
+            return
+
+        pkgs = appman.get_packages(package_type, os, labels=labels)
+        if not pkgs:
+            logger.console("No packages found")
+            return
+
+        for pkg in pkgs:
+            logger.console(pkg.id)
+    except Exception as e:
+        e.verbose = verbose
+        raise
+
+
+@cli.command(cls=BaseCommand, short_help="List packages added by user")
+@click.pass_context
+def list(ctx, package_type, package_id, labels, os, verbose, quiet):
+    """List packages added by user"""
     os = os or ctx.obj["os"]
-    verbose = ctx.obj["verbose"]
     appman = ctx.obj["appman"]
 
     try:
@@ -134,9 +144,9 @@ def list(ctx, package_type, package_id, labels, os):
 
         if package_id:
             if appman.has_user_package(package_type, package_id):
-                util.print_info(f"Package '{package_id}' found")
+                logger.console(f"Package '{package_id}' found")
             else:
-                util.print_info(f"Package '{package_id}' not found")
+                logger.console(f"Package '{package_id}' not found")
             return
 
         pkgs = appman.get_user_packages(package_type, os=os, labels=labels)
@@ -144,25 +154,25 @@ def list(ctx, package_type, package_id, labels, os):
             msg = f"No {package_type} packages found"
             if labels:
                 msg += f" with labels: {', '.join(labels)}"
-            util.print_info(msg)
+            logger.console(msg)
             return
 
         for pkg in pkgs:
             msg = f"  • {pkg.id}"
             if pkg.labels:
                 msg += f" ({', '.join(pkg.labels)})"
-            util.print_info(msg)
+            logger.console(msg)
     except Exception as e:
         e.verbose = verbose
         raise
 
 
-@cli.command(cls=BaseCommand)
+@cli.command(cls=BaseCommand, short_help="Add package to user list")
 @click.option("--interactive", "-i", is_flag=True, help="Enter interactive mode")
 @click.pass_context
-def add(ctx, package_type, package_id, labels, os, interactive):
+def add(ctx, package_type, package_id, labels, os, verbose, quiet, interactive):
+    """Add package to user list"""
     os = os or ctx.obj["os"]
-    verbose = ctx.obj["verbose"]
     appman = ctx.obj["appman"]
 
     try:
@@ -172,13 +182,13 @@ def add(ctx, package_type, package_id, labels, os, interactive):
 
             pkgs = appman.get_packages(package_type, os)
             if not pkgs:
-                util.print_warning(f"No {package_type} package definitions found")
+                logger.warning(f"No {package_type} package definitions found")
                 return
             usr_pkgs = appman.get_user_packages(package_type)
             qname = "add"
             choices = get_user_packages_choices(pkgs, usr_pkgs, qname)
             if not choices:
-                util.print_warning(f"No {package_type} package definitions found")
+                logger.warning(f"No {package_type} package definitions found")
                 return
             questions = get_prompt_questions(
                 "checkbox", f"Select {package_type} packages to add:", qname, choices
@@ -188,36 +198,36 @@ def add(ctx, package_type, package_id, labels, os, interactive):
             for pid in pids:
                 pkg = appman.get_package(package_type, pid)
                 appman.add_user_package(pkg, labels)
-                util.print_success(f"Added {pid} package")
+                logger.success(f"Added {pid} package")
             return
 
         pkg = appman.get_package(package_type, package_id)
         if not pkg:
-            util.print_info(f"Package definition for '{package_id}' not found")
+            logger.warning(f"Package definition for '{package_id}' not found")
             pkgs = appman.get_packages(package_type, os)
             if pkgs:
-                util.print_info(
+                logger.console(
                     f"You can choose from this list: {', '.join([p.id for p in pkgs])}"
                 )
             return
 
         if appman.has_user_package(package_type, package_id):
-            util.print_warning(f"Package '{package_id}' already found")
+            logger.warning(f"Package '{package_id}' already found")
             return
 
         appman.add_user_package(pkg, labels)
-        util.print_success(f"Package '{package_id}' added")
+        logger.success(f"Package '{package_id}' added")
     except Exception as e:
         e.verbose = verbose
         raise
 
 
-@cli.command(cls=BaseCommand)
+@cli.command(cls=BaseCommand, short_help="Remove package from user list")
 @click.option("--interactive", "-i", is_flag=True, help="Enter interactive mode")
 @click.pass_context
-def remove(ctx, package_type, package_id, labels, os, interactive):
+def remove(ctx, package_type, package_id, labels, os, verbose, quiet, interactive):
+    """Remove package from user list"""
     os = os or ctx.obj["os"]
-    verbose = ctx.obj["verbose"]
     appman = ctx.obj["appman"]
 
     try:
@@ -227,15 +237,13 @@ def remove(ctx, package_type, package_id, labels, os, interactive):
 
             usr_pkgs = appman.get_user_packages(package_type, labels=labels)
             if not usr_pkgs:
-                util.print_warning(f"No {package_type} user packages found")
+                logger.warning(f"No {package_type} user packages found")
                 return
             qname = "remove"
             pkgs = appman.get_packages(package_type, os)
             choices = get_user_packages_choices(pkgs, usr_pkgs, qname)
             if not choices:
-                util.print_warning(
-                    f"No matching {package_type} package definitions found"
-                )
+                logger.warning(f"No matching {package_type} package definitions found")
                 return
             questions = get_prompt_questions(
                 "checkbox", f"Select {package_type} packages to remove:", qname, choices
@@ -245,20 +253,20 @@ def remove(ctx, package_type, package_id, labels, os, interactive):
             for pid in pids:
                 usr_pkg = appman.get_user_package(package_type, pid)
                 appman.remove_user_package(usr_pkg)
-                util.print_success(f"Removed {pid} package")
+                logger.success(f"Removed {pid} package")
             return
 
         usr_pkg = appman.get_user_package(package_type, package_id)
         if not usr_pkg:
-            util.print_warning(f"Package '{package_id}' was not found")
+            logger.warning(f"Package '{package_id}' was not found")
             pkgs = appman.get_user_packages(package_type, labels=labels)
             if pkgs:
-                util.print_info(
+                logger.console(
                     f"You can choose from this list: {', '.join([p.id for p in pkgs])}"
                 )
             return
         appman.remove_user_package(usr_pkg)
-        util.print_success(f"Package '{package_id}' removed")
+        logger.success(f"Package '{package_id}' removed")
     except Exception as e:
         e.verbose = verbose
         raise
@@ -300,18 +308,18 @@ def prompt_package_type(suffix="", choices=config.ptchoices()):
     return ptype
 
 
-def run_command(ctx, action, package_id, package_type, labels, os, test, verbose):
+def run_command(
+    ctx, action, package_id, package_type, labels, os, test, verbose, quiet
+):
     os = os or ctx.obj["os"]
     appman = ctx.obj["appman"]
 
     if package_id:
         if not appman.has_user_package(package_type, package_id):
-            util.print_info(
-                f"Package '{package_id}' not found. Make sure to add it first"
-            )
+            logger.info(f"Package '{package_id}' not found. Make sure to add it first")
             return
         package = appman.get_package(package_type, package_id)
-        package_run(package, action, appman, os, test, verbose, idprovided=True)
+        package_run(package, action, appman, os, test, verbose, quiet, idprovided=True)
     else:
         packages = appman.get_user_packages(package_type, os=os, labels=labels)
         if not packages:
@@ -319,44 +327,62 @@ def run_command(ctx, action, package_id, package_type, labels, os, test, verbose
             if labels:
                 msg += f" with labels: {', '.join(labels)}"
             msg += ". Make sure to add them first"
-            util.print_info(msg)
+            logger.info(msg)
             return
 
         for pkg in packages:
             package = appman.get_package(pkg.type, pkg.id)
             if package:
-                package_run(package, action, appman, os, test, verbose)
+                package_run(package, action, appman, os, test, verbose, quiet)
 
 
-def package_run(package, action, appman, os, test, verbose, idprovided=False):
-    formula = appman.find_best_formula(os, package)
+def package_run(package, action, appman, os, test, verbose, quiet, idprovided=False):
+    formula = package.find_best_formula(os, appman.config)
 
     if not formula:
-        util.print_warning(f"Formula not found for {package.name}")
+        logger.warning(f"Formula not found for {package.name}")
         return
 
     if not test:
-        util.print_info(
-            f"{util.get_verb(action, 'present').capitalize()} {package.name}"
-        )
+        logger.info(f"{util.get_verb(action, 'present').capitalize()} {package.name}")
 
-    formula.init(test)
-    result = package.run(formula, action, test=test, verbose=verbose)
+        verifiable = formula.has_command("installed")
+        if verifiable:
+            if action == "install" and package.is_installed(formula):
+                logger.warning(f"{package.name} is already installed")
+                return
+            if action == "uninstall" and not package.is_installed(formula):
+                logger.warning(f"{package.name} is not installed")
+                return
+        elif verbose:
+            logger.warning(f"Not able to verify if {package.name} is installed")
+
+        if not idprovided:
+            formula.init(test)
+
+    result = package.run(formula, action, test=test, verbose=verbose, quiet=quiet)
 
     if test:
         return
 
-    if result.returncode == 0:
-        if idprovided:
-            util.print_success(
-                f"{package.name} {util.get_verb(action, 'past')} successfully"
-            )
-    else:
-        util.print_error(f"{package.name} was not {util.get_verb(action, 'past')}")
+    if result.returncode != 0:
+        logger.error(f"{package.name} was not {util.get_verb(action, 'past')}")
         if result.stderr:
-            util.print_error(util.parse_stmsg(result.stderr))
+            logger.error(util.parse_stmsg(result.stderr))
         if verbose and result.stdout:
-            util.print_info(util.parse_stmsg(result.stdout))
+            logger.info(util.parse_stmsg(result.stdout))
+        return
+
+    if not verifiable:
+        logger.warning(f"Not able to verify if {package.name} was installed")
+        return
+
+    if (action == "install" and package.is_installed(formula)) or (
+        action == "uninstall" and not package.is_installed(formula)
+    ):
+        logger.success(f"{package.name} {util.get_verb(action, 'past')} successfully")
+    else:
+        logger.warning(f"{package.name} was not {util.get_verb(action, 'past')}")
 
 
 def main():
@@ -365,4 +391,4 @@ def main():
     except Exception as e:
         if "verbose" in dir(e) and e.verbose:
             raise
-        util.print_error(e)
+        logger.error(e)
